#!/usr/bin/env python
# coding: utf-8
# Copyright (c) 2015 Fabian Barkhau <fabian.barkhau@gmail.com>
# License: MIT (see LICENSE file)


from __future__ import print_function
from __future__ import unicode_literals


from pycoin.tx.Tx import Tx
from pycoin.serialize import b2h, h2b, b2h_rev, h2b_rev
from . import sanitize
from . import control
from . insight import InsightService # XXX rm when added to next pycoin version


class BtcTxStore(): # TODO use apigen when ported to python 3
    """Bitcoin nulldata output io library."""

    def __init__(self, testnet=False, dryrun=False):
        self.testnet = sanitize.flag(testnet)
        self.dryrun = sanitize.flag(dryrun)
        if self.testnet:
            self.service = InsightService("https://test-insight.bitpay.com/")
        else:
            self.service = InsightService("https://insight.bitpay.com/")

    def writebin(self, rawtx, hexdata):
        """Writes <hexdata> as new nulldata output to <rawtx>."""
        tx = sanitize.unsignedtx(rawtx)
        nulldatatxout = sanitize.nulldatatxout(hexdata)
        tx = control.addnulldata(tx, nulldatatxout)
        return tx.as_hex()

    def readbin(self, rawtx):
        """Returns nulldata from <rawtx> as hexdata."""
        tx = sanitize.tx(rawtx)
        data = control.readnulldata(tx)
        return b2h(data)

    def createtx(self, txins, txouts, locktime="0"):
        """Create unsigned rawtx with given txins/txouts as json data.
        <txins>: '[{"txid" : hexdata, "index" : integer}, ...]'
        <txouts>: '[{"address" : hexdata, "value" : satoshis}, ...]'
        """
        locktime = sanitize.positiveinteger(locktime)
        txins = sanitize.txins(txins)
        txouts = sanitize.txouts(self.testnet, txouts)
        tx = Tx(1, txins, txouts, locktime)
        return tx.as_hex()

    def gettx(self, txid):
        """Returns rawtx from <txid>."""
        txid = sanitize.txid(txid)
        tx = self.service.get_tx(txid)
        return tx.as_hex()

    def signtx(self, rawtx, privatekeys):
        """Sign <rawtx> with  given <privatekeys> as json data.
        <privatekeys>: '["privatekey_in_wif_format", ...]'
        """
        tx = sanitize.tx(rawtx)
        secretexponents = sanitize.secretexponents(self.testnet, privatekeys)
        tx = control.signtx(self.service, self.testnet, tx, secretexponents)
        return tx.as_hex()

    def getutxos(self, addresses):
        """Get current utxos for <address>."""
        addresses = sanitize.addresses(addresses)
        spendables = self.service.spendables_for_addresses(addresses)
        def reformat(spendable):
            return {
                "txid" : b2h_rev(spendable.tx_hash),
                "index" : spendable.tx_out_index,
                "value" : spendable.coin_value,
                "script" : b2h(spendable.script)
            }
        return list(map(reformat, spendables))

    def publish(self, rawtx):
        """Publish signed <rawtx> to bitcoin network."""
        tx = sanitize.signedtx(rawtx)
        if not self.dryrun:
            self.service.send_tx(tx)
        return b2h_rev(tx.hash())

    def createkey(self):
        bip32node = control.createkey(self.testnet)
        return bip32node.wif()

    def store(self, hexdata, privatekeys, changeaddress=None, txouts=None,
              fee="10000", locktime="0"):
        """Store <hexdata> in blockchain and return new txid.
        Utxos taken from <privatekeys> and change sent to <changeaddress>.
        <privatekeys>: '["privatekey_in_wif_format", ...]'
        """
        nulldatatxout = sanitize.nulldatatxout(hexdata)
        secretexponents = sanitize.secretexponents(self.testnet, privatekeys)
        txouts = sanitize.txouts(self.testnet, txouts) if txouts else []
        fee = sanitize.positiveinteger(fee)
        locktime = sanitize.positiveinteger(locktime)
        txid = control.store(self.service, self.testnet, nulldatatxout,
                             secretexponents, changeaddress, txouts, fee, 
                             locktime, publish=(not self.dryrun))
        return b2h_rev(txid)

    def retrieve(self, txid):
        """Returns nulldata stored in blockchain <txid> as hexdata."""
        rawtx = self.gettx(txid)
        return self.readbin(rawtx)




#!/usr/bin/env python
# coding: utf-8
# Copyright (c) 2015 Fabian Barkhau <fabian.barkhau@gmail.com>
# License: MIT (see LICENSE file)


from __future__ import print_function
from __future__ import unicode_literals


import re
from pycoin.key import Key
from pycoin.tx.Tx import Tx
from pycoin.serialize import b2h, h2b, b2h_rev, h2b_rev
from pycoin.tx.script import tools
from pycoin.encoding import bitcoin_address_to_hash160_sec
from pycoin.encoding import wif_to_secret_exponent
from pycoin.tx.pay_to import build_hash160_lookup
from pycoin.tx.TxOut import TxOut
from pycoin.tx.TxIn import TxIn                                                  


class InvalidInput(Exception): pass


def tx(rawtx):
    return Tx.tx_from_hex(rawtx)


def signedtx(rawtx):
    # FIXME validate tx is signed
    return Tx.tx_from_hex(rawtx)


def unsignedtx(rawtx):
    # FIXME validate tx is unsigned
    return Tx.tx_from_hex(rawtx)


def binary(hexdata):
    return h2b(hexdata)


def integer(number):
    return int(number)


def flag(flag):
    return bool(flag)


def positiveinteger(number):
    number = int(number)
    if number < 0:
        raise InvalidInput("Integer may not be < 0!")
    return number


def txid(txhash):
    return h2b_rev(txhash)


def address(address):
    return address # TODO sanitize


def addresses(addresses):
    return list(map(address, addresses))


def txin(txhash, index):
    txhash = txid(txhash)
    index = positiveinteger(index)
    return TxIn(txhash, index)


def txout(testnet, targetaddress, value):
    testnet = flag(testnet)
    targetaddress = address(targetaddress)
    value = positiveinteger(value)
    prefix = b'\x6f' if testnet else b"\0"
    hash160 = b2h(bitcoin_address_to_hash160_sec(targetaddress, prefix))
    script_text = "OP_DUP OP_HASH160 %s OP_EQUALVERIFY OP_CHECKSIG" % hash160
    script_bin = tools.compile(script_text)
    return TxOut(value, script_bin)


def txins(data):
    return list(map(lambda x: txin(x['txid'], x['index']), data))


def txouts(testnet, data):
    return list(map(lambda x: txout(testnet, x['address'], x['value']), data))


def nulldatatxout(hexdata):
    data = binary(hexdata)
    if len(data) > 40:
        raise Exception("Data exceeds maximum of 40 bytes!")
    script_text = "OP_RETURN %s" % b2h(data)
    script_bin = tools.compile(script_text)
    return TxOut(0, script_bin)


def secretexponents(testnet, wifs):
    valid_prefixes = [b'\xef' if testnet else b'\x80']
    return list(map(lambda x: wif_to_secret_exponent(x, valid_prefixes), wifs))





#!/usr/bin/env python
# coding: utf-8
# Copyright (c) 2015 Fabian Barkhau <fabian.barkhau@gmail.com>
# License: MIT (see LICENSE file)


from __future__ import print_function
from __future__ import unicode_literals


import re
import os
from pycoin.ecdsa import public_pair_for_secret_exponent
from pycoin.ecdsa import generator_secp256k1
from pycoin.tx.Tx import Tx
from pycoin.encoding import hash160_sec_to_bitcoin_address
from pycoin.encoding import public_pair_to_hash160_sec
from pycoin.tx.script import tools
from pycoin.serialize import b2h, h2b, b2h_rev, h2b_rev
from pycoin.tx.pay_to import build_hash160_lookup
from pycoin.tx import SIGHASH_ALL 
from pycoin.tx.TxIn import TxIn                                                  
from pycoin.key.BIP32Node import BIP32Node


from . import sanitize


def getnulldata(tx):
    for out in tx.txs_out:
        if re.match("^OP_RETURN", tools.disassemble(out.script)):
            return out
    return None


def addnulldata(tx, nulldatatxout):
    if getnulldata(tx):
        raise Exception("Transaction already has a nulldata output!")
    # TODO validate transaction is unsigned
    tx.txs_out.append(nulldatatxout)
    # TODO validate transaction
    return tx


def readnulldata(tx):
    out = getnulldata(tx)
    if not out:
        return ""
    return h2b(tools.disassemble(out.script)[10:])


def signtx(service, testnet, tx, secretexponents):
    netcode = 'XTN' if testnet else 'BTC'
    lookup = build_hash160_lookup(secretexponents)
    for txin_idx in range(len(tx.txs_in)):
        txin = tx.txs_in[txin_idx]
        utxo_tx = service.get_tx(txin.previous_hash)
        script = utxo_tx.txs_out[txin.previous_index].script
        tx.sign_tx_in(lookup, txin_idx, script, SIGHASH_ALL, netcode=netcode)
    return tx


def findtxins(service, addresses, amount):
    spendables = service.spendables_for_addresses(addresses)
    txins = []
    total = 0
    for spendable in spendables:
        total += spendable.coin_value
        txins.append(TxIn(spendable.tx_hash, spendable.tx_out_index))
        if total >= amount:
            return txins, total
    return txins, total


def secretexponents_to_addresses(testnet, secretexponents):
    prefix = b'\x6f' if testnet else b"\0"
    addresses = []
    for secretexponents in secretexponents:
        public_pair = public_pair_for_secret_exponent(generator_secp256k1, 
                                                      secretexponents)
        hash160 = public_pair_to_hash160_sec(public_pair, compressed=False)
        address = hash160_sec_to_bitcoin_address(hash160, address_prefix=prefix)
        addresses.append(address)
    return addresses


def store(service, testnet, nulldatatxout, secretexponents, 
          changeaddress=None, txouts=None, fee=10000, 
          locktime=0, publish=True):

    # get required satoshis
    txouts = txouts if txouts else []
    required = sum(list(map(lambda txout: txout.coin_value, txouts))) + fee
    
    # get txins
    addresses = secretexponents_to_addresses(testnet, secretexponents)
    txins, total = findtxins(service, addresses, required)
    if total < required:
        msg = "Insufficient funds! Required: %s Available: %s"
        raise Exception(msg % (required, total))

    # setup txouts
    changeaddress = changeaddress if changeaddress else addresses[0]
    changeout = sanitize.txout(testnet, changeaddress, total - required)
    txouts = txouts + [nulldatatxout, changeout]

    # create, sign and publish tx
    tx = Tx(1, txins, txouts, locktime)
    tx = signtx(service, testnet, tx, secretexponents)
    if publish:
        service.send_tx(tx)
    return tx.hash()


def createkey(testnet):
    netcode = 'XTN' if testnet else 'BTC'
    return BIP32Node.from_master_secret(os.urandom(64), netcode=netcode)





#!/usr/bin/env python
# coding: utf-8
# Copyright (c) 2015 Fabian Barkhau <fabian.barkhau@gmail.com>
# License: MIT (see LICENSE file)


import unittest
from btctxstore import BtcTxStore


unsignedrawtx = "0100000001ef5b2eb4b9b0c10449cbabedca45709135d457a04dabd33c1068aaf86562a72b0200000000ffffffff0240420f00000000001976a91491cc9812ca45e7209ff9364ce96527a7c49f1f3188ac3e770400000000001976a9142e330c36e1d0f199fd91446f2210209a0d35caef88ac00000000"


class TestIO(unittest.TestCase):

    def setUp(self):
        self.api = BtcTxStore(dryrun=True, testnet=True)

    def test_readwrite(self):
        rawtx = self.api.writebin(unsignedrawtx, "f483")
        data = self.api.readbin(rawtx)
        self.assertEqual(data, "f483")

    def test_only_one_nulldata_output(self):
        def callback():
            rawtx = self.api.writebin(unsignedrawtx, "f483") # first nulldata
            self.api.writebin(rawtx, "f483") # writing second fails
        self.assertRaises(Exception, callback)

    def test_max_data(self): # TODO move to test/sanitize.py
        max_data = 40 * "aa" # fourty bytes
        self.api.writebin(unsignedrawtx, max_data)
        def callback():
            over_max_data = 41 * "aa" # fourty bytes
            self.api.writebin(unsignedrawtx, over_max_data)
        self.assertRaises(Exception, callback)


class TestCreateTx(unittest.TestCase):

    def setUp(self):
        self.api = BtcTxStore(dryrun=True, testnet=True)

    def test_create(self):
        txins = [
            {
                "txid" : "33a184cba7cf96cc167780561d261201cef0fb424d6591588d88e61bcfdd09c8",
                "index" : 0
            },
            {
                "txid" : "d2f4411e0b29d0f2f4fed66d0e19cef337b2d8fe1a809e4c46833429f6ea87d0",
                "index" : 1
            }
        ]
        txouts = [
            {
                "address" : "migiScBNvVKYwEiCFhgBNGtZ87cdygtuSQ",
                "value" : "750000"
            },
            {
                "address" : "mkSWUYy3ggmbfGMf4PrjKj4LdU45Ytt2DN",
                "value" : "1000104"
            }
        ]
        locktime = 0
        testnet = True
        result = self.api.createtx(txins, txouts, locktime)
        expected = "0100000002c809ddcf1be6888d5891654d42fbf0ce0112261d56807716cc96cfa7cb84a1330000000000ffffffffd087eaf6293483464c9e801afed8b237f3ce190e6dd6fef4f2d0290b1e41f4d20100000000ffffffff02b0710b00000000001976a91422c0f934b5346bd3e14dd47c2eb26c4bdf15eab988aca8420f00000000001976a91436016996a73708c5faa17ac9b76ec380941e545a88ac00000000"
        self.assertEqual(result, expected)


class TestGetTx(unittest.TestCase):

    def setUp(self):
        self.api = BtcTxStore(dryrun=True, testnet=True)

    def test_gettx(self):
        txid = "987451c344c504d07c1fa12cfbf84b5346535da5154006f6dc8399a8fae127eb"
        expected = "0100000001f2963731aa6b1e27a2f94d79620fad643e9e65741b16dadd4e77489c34f20ae2010000008a473044022054d89d41e9a3df9ea6ac367e0a062bd7c65c50232ff36d1f9287cd1851c7444e022069736675af9b94340d64b2b67913c07c4fc2f24006e77ad4df5a13a26e1200350141040319ffdcba35ef3d2577cdf6f07483f4b30865f695d366f02926db1ddd0c03544150b65124baf42601945d1c848bca7970cfa29f538f4ad8cd2564b8f80bb10cffffffff020000000000000000046a02f48370811201000000001976a914f4131906b10615a61af347c56f1223ddc214f95c88ac00000000"
        result = self.api.gettx(txid)
        self.assertEqual(result, expected)


class TestGetUtxos(unittest.TestCase):

    def setUp(self):
        self.api = BtcTxStore(dryrun=True, testnet=True)

    def test_getutxos(self):
        address = "n3mW3o8XNMyH6xHWBkN98rm7zxxxswzpGM"
        expected = [{
            "index": 1, 
            "txid": "987451c344c504d07c1fa12cfbf84b5346535da5154006f6dc8399a8fae127eb",
            "script": "76a914f4131906b10615a61af347c56f1223ddc214f95c88ac",
            "value": 17990000
        }]
        result = self.api.getutxos([address])
        self.assertEqual(result, expected)


class TestSignTx(unittest.TestCase):

    def setUp(self):
        self.api = BtcTxStore(dryrun=True, testnet=True)

    def test_signtx(self):
        txins = [{
          "index": 1, 
          "txid": "987451c344c504d07c1fa12cfbf84b5346535da5154006f6dc8399a8fae127eb"
        }]
        txouts = [{
          "address" : "n3mW3o8XNMyH6xHWBkN98rm7zxxxswzpGM",
          "value" : 17980000
        }]
        privatekeys = ["92JATRBTHRGAACcJb41dAGnh7kQ1wev27tcYWcGA2RZeUJLCcZo"]
        rawtx = self.api.createtx(txins, txouts)
        rawtx = self.api.writebin(rawtx, "f483")
        result = self.api.signtx(rawtx, privatekeys)
        expected = "0100000001eb27e1faa89983dcf6064015a55d5346534bf8fb2ca11f7cd004c544c3517498010000008b4830450221008326d0d915dd8d3f9bcced2d774f4a898ed2c4a4929a06c7539500ded89e92db02201542ce4beda2eb1cfdefa3d647481a2a2f38231c953bba8efbb860fe1f49981d0141040319ffdcba35ef3d2577cdf6f07483f4b30865f695d366f02926db1ddd0c03544150b65124baf42601945d1c848bca7970cfa29f538f4ad8cd2564b8f80bb10cffffffff02605a1201000000001976a914f4131906b10615a61af347c56f1223ddc214f95c88ac0000000000000000046a02f48300000000"
        self.assertEqual(result, expected)


class TestStore(unittest.TestCase):

    def setUp(self):
        self.api = BtcTxStore(dryrun=True, testnet=True)

    def test_store(self):
        privatekeys = ["92JATRBTHRGAACcJb41dAGnh7kQ1wev27tcYWcGA2RZeUJLCcZo"]
        result = self.api.store("f483", privatekeys)
        expected = "6a7311a49b4e59dd3bfcaea75a114d1c3f9cb2e4dbb9b3ed99eef5846a8e1a2a"
        self.assertEqual(result, expected)

    def test_store_txouts(self):
        privatekeys = ["92JATRBTHRGAACcJb41dAGnh7kQ1wev27tcYWcGA2RZeUJLCcZo"]
        txouts = [{
          "address" : "mgBJ5bG9mQw8mHHcVEJghMamQEXeNLtvpt",
          "value" : 10000000
        }]
        result = self.api.store("f483", privatekeys, txouts=txouts)
        expected = "9a015258e681e361d99c763aae09776f32b74cc021fb0151ec43fb4f72a65ef8"
        self.assertEqual(result, expected)

    def test_store_changeaddress(self):
        privatekeys = ["92JATRBTHRGAACcJb41dAGnh7kQ1wev27tcYWcGA2RZeUJLCcZo"]
        changeaddress = "mgBJ5bG9mQw8mHHcVEJghMamQEXeNLtvpt"
        result = self.api.store("f483", privatekeys, changeaddress)
        expected = "186d1f6f97b56249ba9d9b13f2f50642b19877c356879888dde7ff160fae7d80"
        self.assertEqual(result, expected)


class TestRetrieve(unittest.TestCase):

    def setUp(self):
        self.api = BtcTxStore(dryrun=True, testnet=True)

    def test_retrieve(self):
        txid = "987451c344c504d07c1fa12cfbf84b5346535da5154006f6dc8399a8fae127eb"
        result = self.api.retrieve(txid)
        self.assertEqual(result, "f483")


if __name__ == '__main__':
    unittest.main()





#!/usr/bin/env python
# coding: utf-8
# Copyright (c) 2015 Fabian Barkhau <fabian.barkhau@gmail.com>
# License: MIT (see LICENSE file)


import os
from setuptools import setup, find_packages


THISDIR = os.path.dirname(os.path.abspath(__file__))
os.chdir(THISDIR)


VERSION = open("version.txt").readline().strip()
DOWNLOAD_BASEURL = "https://pypi.python.org/packages/source/a/btctxstore/"
DOWNLOAD_URL = DOWNLOAD_BASEURL + "btctxstore-%s.tar.gz" % VERSION


setup(
    name='btctxstore',
    version=VERSION,
    description=('A simple library to store/retrieve information '
                 'in bitcoin transactions using OP_RETURN.'),
    long_description=open("README.rst").read(),
    keywords=("Bitcoin, OP_RETURN, store, tx, transaction"),
    url='https://github.com/F483/btctxstore/',
    author='Fabian Barkhau',
    author_email='fabian.barkhau@gmail.com',
    license='MIT',
    packages=find_packages(),
    download_url = DOWNLOAD_URL,
    test_suite="tests.api",
    install_requires=[
        'pycoin == 0.52'
    ],
    tests_require=[
      'ipython',
      'pylint',
      'pudb' # import pudb; pu.db # set break point
    ],
    zip_safe=False
)


